"""
    Dummy conftest.py for carg_io.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

import pytest
from carg_io.core import Parameter, ParameterSet, units, NaN
from carg_io.spaces import Space
import numpy as np

class Box(ParameterSet):
    Length:Parameter = 1 * units.meter
    Width:Parameter = 1* units.meter
    Height:Parameter = 1* units.meter


class Block(ParameterSet):
    Length:Parameter = 1 * units.meter
    Width:Parameter = 1 * units.meter
    Height:Parameter = 1 * units.meter
    Density:Parameter = 2 * units.kilogram / units.meter**3

    @property
    def Volume(self) -> Parameter:
        l = self.Length['m']
        w = self.Width['m']
        h = self.Height['m']
        return Parameter('Volume', l*w*h * units.meter**3)

    @property
    def Mass(self) -> Parameter:
        v = self.Volume['m**3']
        rho = self.Density['kg/m**3']
        return Parameter('Mass', v*rho * units.kg)

    @property
    def LengthWidthRatio(self) -> Parameter:
        l = self.Length['m']
        w = self.Width['m']
        return Parameter('LengthWidthRatio', l/w * units.dimensionless)

@pytest.fixture
def block():
    return Block()


@pytest.fixture
def block_space(block):
    space = Space(Block)
    space.expand(Block.Length, 'm', np.linspace(1,10,10))
    space.expand(Block.Width, 'm', np.linspace(1,10,10))
    space.expand(Block.Height, 'm', np.linspace(1,10,10))
    return space


class SpecializedBlock(Block):
    FilledPercentage: Parameter = 100 * units.dimensionless

    @property
    def Mass(self) -> Parameter:
        v = self.Volume['m**3']
        rho = self.Density['kg/m**3']
        f = self.FilledPercentage[None]
        return Parameter('Mass', v*rho*f * units.kg)

@pytest.fixture
def specialized_block():
    return SpecializedBlock()


import pytest
from carg_io.core import Parameter, ParameterSet, units, NaN
from carg_io.spaces import Space
from conftest import block_space, Block
import numpy as np


def test_space(block_space):
    assert len(block_space) == 10**3

def test_criteria(block_space):
    block_space.add_criteria("Volume", 'm**3', lambda v: v < 10*10*9)
    # There should be four cases that will not pass this criteria:
    # everything 10 m, or any combination of two parameters being 10 m.
    # = 4 cases
    assert len(block_space) == 10**3-4
    

def test_init_type_error():
    instance = Block()
    with pytest.raises(TypeError):
        s = Space(instance)


import pytest
from carg_io.core import Parameter, ParameterSet, units, NaN
from conftest import Box, block, specialized_block
from carg_io.core import SettingAttributeNotAllowed




def test_parameter1():
    
    height = Parameter('height', 8*units.meter)
    assert height.is_default
    assert height.name == 'height'
    assert height['mm'] == 8000


def test_parameter_set():

    box = Box()
    assert box.Length['km'] == 1e-3

def test_parameter_get_dimensionless(block):
    assert isinstance(block.LengthWidthRatio[None], float)
    assert isinstance(block.LengthWidthRatio[:], float)

# def test_hash():
#     box = Box()
#     box.Height['m'] = 99
#     identical_box = Box()
    
#     identical_box.Height['mm'] = 99_000
#     hash1 = hash(box)
#     hash2 = hash(identical_box)
#     assert hash1 == hash2 # the hash checks if the instance values are equal
#     assert box != identical_box # the instances themselves are not equal

def test_pickle():
    box = Box()
    file = box.to_pickle()

def test_parm_equality1():
    box1 = Box()
    box2 = Box()
    assert box1.Length == box2.Length

def test_parm_equality2():
    box1 = Box()
    box2 = Box()
    box1.Length["cm"] == box1.Length["m"]/100
    assert box1.Length == box2.Length

def test_set_equality():
    box1 = Box()
    box2 = Box()
    assert box1 == box2

def test_set_not_equality():
    box1 = Box()
    box2 = Box()
    box2.Length["m"] = 123
    assert not box1 == box2

def test_setting_parm_error():
    box1 = Box()
    with pytest.raises(SettingAttributeNotAllowed):
        box1.Length = 123

def test_dependent_parm(block):
    assert block.Mass["pounds"]

def test_default_value(block):
    """Test that setting a value changes the 'is_default' attribute
    from True to False."""
    assert block.Length.is_default
    block.Length["m"] = block.Length["m"] + 1
    assert not block.Length.is_default

def test_repr(block):
    assert block.__repr__() == "ParameterSet(name='Block')"


def test_inheritance(specialized_block):
    assert isinstance(specialized_block.FilledPercentage[None], (int, float))
    specialized_block.FilledPercentage[None] = 0
    assert specialized_block.Mass["kg"] == 0


# from __future__ import annotations
# from typing import List
# from .core import ParameterSet
# import zipfile
# from pathlib import Path
# import tkinter as tk

# class Container:
#     """
#     A collection of i/o parameters and any result files.

#     This is an object-oriented datastorage, and propably pretty slow.
#     But for more intensive stuff, we might have a relational database equivalent

#     """

#     i:List[ParameterSet]
#     o:List[ParameterSet]

#     def __init__(self, input:List[ParameterSet]=[], output:List[ParameterSet]=[]):
#         self.input = input
#         self.output = output
#         self.additional_files = []
    
#     def append_file(self, file:Path|str):
#         self.additional_files.append(Path(file))

#     def append_input(self, input:ParameterSet):
#         self.input.append(input)
    
#     def append_output(self, output:ParameterSet):
#         self.output.append(output)

#     @staticmethod
#     def load(file:str, subclass:Container):
#         input, output = [], []
#         idict = {i.name:i for i in subclass.i}
#         odict = {o.name:o for o in subclass.o}
#         with zipfile.ZipFile(file, mode="r") as archive:
#             for file in archive.filelist:
#                 data = archive.read(file.filename)
#                 name, _ = file.filename.split('.')
                
#                 if name in idict:
#                     cls = idict[name]
#                     inst = cls.from_serial_pickle(data)
#                     input.append(inst)
#                 elif name in odict:
#                     cls = odict[name]
#                     inst = cls.from_serial_pickle(data)
#                     output.append(inst)
#                 else:
#                     pass #TODO
#         return subclass(input=input, output=output)

#     def save(self, file:Path):
#         with zipfile.ZipFile(file, mode="w") as archive:
#             for filename in self.additional_files:
#                 archive.write(filename)
            
#             for pset in self.input + self.output:
#                 filename = pset.to_pickle()
#                 archive.write(filename)
            
#     def file_ui(self):
#         root = tk.Tk()
        
#         input_column = tk.LabelFrame(root, text='INPUT', font=('Helvetica 10 bold'))
#         sbi = tk.Scrollbar(input_column, orient = 'vertical')
#         sbi.pack(side=tk.LEFT, fill=tk.Y)
#         input_parm = tk.Frame(input_column, width=360, height=520)
#         for inp in self.input:
#             frame, entries = inp._to_tk(input_parm, state='disabled')
#             frame.pack(padx=(10,10), pady=(10,10), side=tk.TOP)
        

#         output_column = tk.LabelFrame(root, text='OUTPUT', font=('Helvetica 10 bold'))
#         output_parm = tk.Frame(output_column, width=360, height=520)
#         for inp in self.output:
#             inp:ParameterSet
#             frame, entries = inp._to_tk(output_parm, state='disabled')
#             frame.pack(padx=(10,10), pady=(10,10), side=tk.TOP)
        
#         input_parm.pack(side=tk.TOP)
#         input_column.pack(side=tk.LEFT)
#         output_parm.pack(side=tk.TOP)
#         output_column.pack(side=tk.LEFT)
        

#         root.mainloop()


#     def get_partial_hash(self):
#         pass
    
#     def __hash__(self):
#         pass

#     def open(self):
#         pass


    

# class Vessel:
#     """
#     A collection of containers

#     """

#     def __init__(self, containers:List[Container]):
#         self.containers = containers

    

# class Database:
#     """
    
#     """
#     pass